use std::collections::HashMap;

use ldap3::{Ldap, LdapConnAsync, LdapConnSettings, Scope, SearchEntry};
use serde::{Deserialize, Serialize};
use ts_rs::TS;
use utoipa::ToSchema;

use crate::{
    config::{Configuration, LDAP},
    errors::ApiError,
};

#[derive(Debug, Clone, Serialize, Deserialize, TS, ToSchema)]
#[ts(export)]
pub struct ActiveDirectoryUser {
    pub name: String,
    pub surname: String,
    pub email: String,
    pub id: String,
}

pub struct LDAPBackend {
    connection: Ldap,
    config: LDAP,
    bound: bool,
}

impl LDAPBackend {
    pub async fn from_config(config: &Configuration) -> Result<Self, ApiError> {
        // create connection to LDAP server
        let ldap_settings = LdapConnSettings::new().set_no_tls_verify(config.ldap.skip_tls_verify);
        let (connection, ldap) = LdapConnAsync::with_settings(ldap_settings, &config.ldap.server)
            .await
            .map_err(|e| ApiError::InternalError(format!("LDAP Server connection: {e}")))?;
        ldap3::drive!(connection);

        Ok(Self {
            config: config.ldap.clone(),
            connection: ldap,
            bound: false,
        })
    }

    pub async fn unbind(&mut self) {
        let _ = self.connection.unbind().await;
        self.bound = false;
    }

    pub async fn ad_bind(&mut self, user: &str, password: &str) -> Result<(), ApiError> {
        self.connection
            .simple_bind(&format!("{}\\{user}", self.config.ad_domain), &password)
            .await
            .map_err(|_e| ApiError::InvalidCredentials)?;

        self.bound = true;

        Ok(())
    }

    pub async fn get_user_detail(
        &mut self,
        user_id: &str,
    ) -> Result<HashMap<String, Vec<String>>, ApiError> {
        // check if elevated user has to be used
        if self.config.elevated_search {
            let id = self.config.elevated_user_id.clone();
            let password = self.config.elevated_user_pw.clone();

            self.ad_bind(&id, &password).await?;
        }
        let (search_result, _) = self
            .connection
            .search(
                &self.config.user_search_base,
                Scope::Subtree,
                // "(objectClass=*)",
                &format!("(&(ObjectCategory=Person)(sAMAccountName={user_id}))"),
                vec!["givenName", "sn", "mail"],
                // vec!["*"],
            )
            .await
            .map_err(|e| ApiError::InternalError(format!("LDAP search error: {e}")))?
            .success()
            .map_err(|e| ApiError::InternalError(format!("LDAP search error: {e}")))?;

        let attributes = if let Some(entry) = search_result.first().cloned() {
            SearchEntry::construct(entry).attrs
        } else {
            HashMap::new()
        };

        Ok(attributes)
    }

    pub async fn get_ad_user_list(&mut self) -> Result<Vec<ActiveDirectoryUser>, ApiError> {
        // check if elevated user has to be used
        if self.config.elevated_search {
            let id = self.config.elevated_user_id.clone();
            let password = self.config.elevated_user_pw.clone();

            self.ad_bind(&id, &password).await?;
        }

        let (search_result, _) = self
            .connection
            .search(
                &self.config.user_search_base,
                Scope::Subtree,
                &format!("(&(ObjectCategory=Person))"),
                vec!["sAMAccountName", "givenName", "sn", "mail"],
            )
            .await
            .map_err(|e| ApiError::InternalError(format!("LDAP search error: {e}")))?
            .success()
            .map_err(|e| ApiError::InternalError(format!("LDAP search error: {e}")))?;

        let result = search_result
            .into_iter()
            .map(|entry| {
                let user_info = SearchEntry::construct(entry).attrs;

                ActiveDirectoryUser {
                    name: user_info
                        .get("givenName")
                        .map(|v| v.first().cloned())
                        .flatten()
                        .unwrap_or_default(),
                    surname: user_info
                        .get("sn")
                        .map(|v| v.first().cloned())
                        .flatten()
                        .unwrap_or_default(),
                    email: user_info
                        .get("mail")
                        .map(|v| v.first().cloned())
                        .flatten()
                        .unwrap_or_default(),
                    id: user_info
                        .get("sAMAccountName")
                        .map(|v| v.first().cloned())
                        .flatten()
                        .unwrap_or_default(),
                }
            })
            .collect();

        Ok(result)
    }
}

#[cfg(test)]
mod test {
    use ldap3::{exop::WhoAmI, LdapConnAsync, LdapConnSettings, Scope, SearchEntry};

    #[tokio::test]
    async fn test_ldap() {
        let (connection, mut ldap) = LdapConnAsync::with_settings(
            LdapConnSettings::new().set_no_tls_verify(true),
            "ldaps://localhost:636",
        )
        .await
        .unwrap();
        ldap3::drive!(connection);

        // bind to server
        let res = ldap
            // .simple_bind("cn=read-only-admin,dc=example,dc=com", "password")
            .simple_bind(
                // "CN=Abel Austin,OU=Accounting,OU=Mylab Users,DC=mylab,DC=local",
                // "MYLAB\\A0H67123",
                // "cn=read-only-admin,dc=example,dc=com",
                // "DEV-AD\\Administrator",
                // "admin123!",
                "DEV-AD\\einstein",
                "einstein",
            )
            .await;

        println!("{res:?}");

        let (res, re) = ldap
            .search(
                // "CN=Abel Austin,OU=Accounting,OU=Mylab Users,DC=mylab,DC=local",
                "OU=DevUsers,DC=example,DC=org",
                Scope::Subtree,
                "(objectClass=*)",
                // "(&(ObjectCategory=Person)(sAMAccountName=A0H67123))",
                vec!["givenName", "sn"],
            )
            .await
            .unwrap()
            .success()
            .unwrap();

        for entry in res {
            println!("{:?}", SearchEntry::construct(entry));
        }

        let test = ldap.extended(WhoAmI).await.unwrap().0;
        println!("{test:?}");
        // let whoami: ldap3::exop::WhoAmIResp = test.parse();
        // println!("{whoami:?}");

        let _ = ldap.unbind().await;
    }
}
