use std::{future::IntoFuture, net::Ipv4Addr, time::Duration};

use axum::http::HeaderName;
use axum_jwt_login::{AuthSessionWithApiKeyBuilder, JWTKeyPair};
use backend::ApiBackend;
use error_stack::{Report, ResultExt};
use routes::api_keys::models::ApiKey;
use sqlx::PgPool;
use tokio::net::TcpListener;
use tokio_util::sync::CancellationToken;
use tracing::info;

use crate::APP_NAME;
use crate::{config::Configuration, errors::AppError};

pub mod backend;
mod description;
pub mod routes;

pub async fn start(
    config: &Configuration,
    cancellation_token: CancellationToken,
) -> Result<(), Report<AppError>> {
    // Initialize database pool
    let pool = PgPool::connect_lazy(&config.database_url).change_context(AppError)?;
    // Get API Keys from database
    let api_keys = ApiKey::get_all_with_secret_attached(&pool, config)
        .await
        .change_context(AppError)?;

    // Initialize API Backend
    let backend = ApiBackend::new(pool.clone(), config);
    // Create Session storage for API
    let session = AuthSessionWithApiKeyBuilder::new()
        .backend(backend)
        .api_header_name(HeaderName::from_static("x-api-key"))
        .initial_api_keys(api_keys)
        .jwt_key_pair(JWTKeyPair::from_secret(&config.token_secret))
        .session_length(Duration::from_secs(config.session_length))
        .build()
        .change_context(AppError)?;

    // Create routes
    let router = routes::create_routes(session);

    // Define listener
    let port = config.port;
    let listener = TcpListener::bind((Ipv4Addr::LOCALHOST, port))
        .await
        .change_context(AppError)?;

    // Serve API
    let server = axum::serve(listener, router.into_make_service())
        .with_graceful_shutdown(shutdown_signal(cancellation_token));

    tokio::spawn(server.into_future());

    info!("{APP_NAME} up and running on port '{port}'.");

    Ok(())
}

async fn shutdown_signal(stop_signal: CancellationToken) {
    // Wait for shutdown event.
    stop_signal.cancelled().await;

    info!("Shutting down {APP_NAME}...");
}

// // Set Report Colour Mode to NONE
// Report::set_color_mode(error_stack::fmt::ColorMode::None);

// // Enable the `INFO` level for anything in `darl` as default
// let level_filter =
//     filter::Targets::new().with_target(env!("CARGO_PKG_NAME"), DEFAULT_LOG_LEVEL_FILTER);
// let (level_filter, tracing_target_reload_handle) = reload::Layer::new(level_filter);

// // Prepare logging to file
// let file_appender = RollingFileAppenderBase::new(
//     ROOT_PATH.with_file_name(format!("{}.log", env!("CARGO_PKG_NAME"))),
//     RollingConditionBase::new().max_size(1024 * 1024 * 2),
//     5,
// )
// .change_context(ServiceError::Starting)?;
// let (non_blocking, _guard) = tracing_appender::non_blocking(file_appender);

// // Prepare live logging to config webserver
// let (log_receiver, live_log_layer) = LiveLogLayer::new();

// // prepare initialization of logging
// let log_layers = tracing_subscriber::registry()
//     .with(level_filter)
//     // .with(filter::LevelFilter::DEBUG)
//     .with(live_log_layer)
//     .with(
//         fmt::Layer::default()
//             .with_target(false)
//             .with_ansi(false)
//             .with_writer(non_blocking),
//     );

// // also log to console in debug mode
// #[cfg(debug_assertions)]
// let stdout_log = tracing_subscriber::fmt::layer().pretty();
// #[cfg(debug_assertions)]
// let log_layers = log_layers.with(stdout_log);

// // Initialize logging
// log_layers.init();

// // Initialize local database
// let local_database = LocalDatabase::init()
//     .await
//     .change_context(ServiceError::Starting)?;

// // Load configuration from config files
// let (config, external_database) = Configuration::initialize(&local_database)
//     .await
//     .change_context(ServiceError::Starting)?;
// let standalone_external_db = StandaloneExternalDatabase::from(&external_database);

// // change log level to configured value
// if let Err(error) = tracing_target_reload_handle.modify(|filter| {
//     *filter = filter::Targets::new().with_target(env!("CARGO_PKG_NAME"), &config.log_level)
// }) {
//     error!("{error}");
// }

// // prepare and start connections
// let connections = MachineConnections::init(&local_database, &standalone_external_db, &config)
//     .await
//     .change_context(ServiceError::Starting)?;

// // start config server
// ConfigServer::start(
//     config.extended_config.webserver.config_server_port,
//     config.opc_configuration.clone(),
//     &local_database,
//     standalone_external_db,
//     connections,
//     tracing_target_reload_handle,
//     log_receiver,
// )
// .await;

// // start webserver
// WebServer::start(&config, &local_database).await;

// // initialize Logging to external database
// external_database.start_writer().await;

// info!("{APP_NAME} service is now running...");

// // block thread
// loop {
//     // Poll shutdown event.
//     if (stop_signal.recv().await).is_some() {
//         // Break the loop either upon stop or channel disconnect
//         info!("Shutting down {APP_NAME} service");
//         break;
//     };
// }

// Ok(())
