use std::{
    fmt::Display,
    hash::{DefaultHasher, Hash, Hasher},
};

use chrono::NaiveDateTime;
use minisign::{SecretKey, SecretKeyBox};
use serde::{Deserialize, Serialize};
use sqlx::PgPool;
use ts_rs::TS;
use utoipa::ToSchema;

use crate::{api::routes::signature::sql, errors::ApiError};

#[derive(Debug, Clone, Serialize, Deserialize, TS, ToSchema)]
#[ts(export)]
pub struct SigningKeyStatus {
    pub configured: bool,
    pub verified: Option<bool>,
}

pub struct KeyData {
    pub key: Option<String>,
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct PublicKeyID {
    pub timestamp: NaiveDateTime,
    pub user_id: String,
}

#[derive(Debug, Clone, Deserialize, TS, ToSchema)]
#[ts(export)]
pub struct SigningKeyRequest {
    pub secret: String,
}

impl SigningKeyRequest {
    pub async fn to_private_key(
        self,
        pool: &PgPool,
        user_id: &str,
    ) -> Result<Option<SecretKey>, ApiError> {
        // try to get private key from database
        let private_key = match sql::get_private_key(pool, user_id).await?.key {
            Some(key) => key,
            None => return Ok(None),
        };

        // create MinSign Box from key string
        let private_key_box = SecretKeyBox::from_string(&private_key)?;
        // and the box can be opened using the password to reveal the original secret key:
        let secret_key = private_key_box.into_secret_key(Some(self.secret))?;

        Ok(Some(secret_key))
    }
}

#[derive(Debug, Clone, Deserialize, Serialize, TS, ToSchema)]
#[ts(export)]
pub struct VerificationRequest {
    pub signature: String,
    pub signed_from: String,
}

#[derive(Debug, Hash)]
pub struct SignatureTest {
    pub value1: String,
    pub value2: bool,
    pub value3: Vec<String>,
}

impl SignatureTest
where
    Self: Hash,
{
    pub fn calculate_hash(&self) -> u64 {
        let mut s = DefaultHasher::new();
        self.hash(&mut s);
        s.finish()
    }
}
