use axum::{debug_handler, Extension, Json};
use axum_jwt_login::UserPermissions;
use models::User;
use permissions::{Permission, PermissionDetail};
use utoipa_axum::{router::OpenApiRouter, routes};

use crate::{
    api::{
        backend::{
            ldap::{ActiveDirectoryUser, LDAPBackend},
            ApiBackend,
        },
        description::USERS_TAG,
        routes::auth::models::Credentials,
    },
    errors::ApiError,
    permission_required,
    utils::create_random,
};

pub mod models;
pub mod permissions;
pub mod sql;

// expose the OpenAPI to parent module
pub fn router() -> OpenApiRouter {
    let read = OpenApiRouter::new()
        .routes(routes!(get_users))
        .route_layer(permission_required!(Permission::Read(
            PermissionDetail::Users
        )));
    let write = OpenApiRouter::new()
        .routes(routes!(create_user, update_user))
        .routes(routes!(get_ad_users))
        .route_layer(permission_required!(Permission::Write(
            PermissionDetail::Users
        )));

    OpenApiRouter::new().merge(read).merge(write)
}

#[debug_handler]
#[utoipa::path(
    get,
    path = "/users",
    summary = "Get all Users",
    description = "Get a list of all users.",
    responses(
        (status = OK, body = Vec<User>, description = "List of users"),
    ),
    security(
        ("user_auth" = ["read:users",]),
    ),
    tag = USERS_TAG)]
pub async fn get_users(
    Extension(backend): Extension<ApiBackend>,
) -> Result<Json<Vec<User>>, ApiError> {
    Ok(Json(sql::get_users(backend.pool(), None, None).await?))
}

#[debug_handler]
#[utoipa::path(
    put,
    path = "/users",
    summary = "Change User details",
    description = "Update user information / permissions / groups ",
    request_body(content = User, description = "User details", content_type = "application/json"),
    responses(
        (status = OK, description = "User successfully updated"),
    ),
    security(
        ("user_auth" = ["write:users"]),
    ),
    tag = USERS_TAG)]
pub async fn update_user(
    Extension(backend): Extension<ApiBackend>,
    Json(user): Json<User>,
) -> Result<(), ApiError> {
    sql::update_user(backend.pool(), &user).await?;

    Ok(())
}

#[debug_handler]
#[utoipa::path(
    post,
    path = "/users",
    summary = "Create a new User",
    description = "Creates a new user with the given information ",
    request_body(content = User, description = "User details", content_type = "application/json"),
    responses(
        (status = OK, description = "User successfully created (Assigned Password in body)", body = String),
    ),
    security(
        ("user_auth" = ["write:users",]),
    ),
    tag = USERS_TAG)]
pub async fn create_user(
    Extension(backend): Extension<ApiBackend>,
    Json(user): Json<User>,
) -> Result<String, ApiError> {
    // create password if not Active Directory user
    let (password, hash) = match user.active_directory_auth {
        true => (String::new(), None),
        false => {
            let salt = create_random(20);
            let argon_config = argon2::Config::default();

            let password = create_random(10);
            let password_hash =
                argon2::hash_encoded(password.as_bytes(), salt.as_bytes(), &argon_config)?;

            (password, Some(password_hash))
        }
    };
    // create user
    sql::create_new_user(backend.pool(), &user, hash).await?;
    // send created password back to frontend
    Ok(password)
}

#[debug_handler]
#[utoipa::path(
    get,
    path = "/users/available_ad_users",
    summary = "Get Active Directory Users",
    description = "Get all Available Users from the Active Directory that are not already registered with this API.",
    responses(
        (status = OK, body = Vec<ActiveDirectoryUser>, description = "List of AD users"),
    ),
    security(
        ("user_auth" = ["write:users",]),
    ),
    tag = USERS_TAG)]
pub async fn get_ad_users(
    Extension(backend): Extension<ApiBackend>,
    Json(credentials): Json<Option<Credentials>>,
) -> Result<Json<Vec<ActiveDirectoryUser>>, ApiError> {
    let api_user_ids: Vec<String> = sql::get_users(backend.pool(), None, None)
        .await?
        .iter()
        .map(|user| user.id())
        .collect();
    let mut ldap = LDAPBackend::from_config(backend.config()).await?;
    // bind to AD user if credentials are given
    if let Some(credentials) = credentials {
        ldap.ad_bind(&credentials.id, &credentials.password).await?;
    }
    let ad_users = ldap
        .get_ad_user_list()
        .await?
        .into_iter()
        .filter(|entry| !api_user_ids.contains(&entry.id))
        .collect();
    // disconnect from AD server
    ldap.unbind().await;

    Ok(Json(ad_users))
}
