use std::path::PathBuf;

use config::{Config, Environment, File};
use error_stack::{Report, ResultExt};
use serde::Deserialize;

use crate::errors::AppError;

fn default_port() -> u16 {
    8080
}

#[derive(Debug, Deserialize, Default, Clone)]
#[allow(unused)]
pub struct ConfigInfo {
    pub location: Option<String>,
}

#[derive(Debug, Deserialize, Default, Clone)]
pub struct LDAP {
    pub server: String,
    #[serde(default)]
    pub skip_tls_verify: bool,
    #[serde(default)]
    pub ad_domain: String,
    #[serde(default)]
    pub user_search_base: String,
    #[serde(default)]
    pub elevated_search: bool,
    #[serde(default)]
    pub elevated_user_id: String,
    #[serde(default)]
    pub elevated_user_pw: String,
}

#[derive(Debug, Deserialize, Default, Clone)]
#[allow(unused)]
pub struct Configuration {
    #[serde(default)]
    pub config: ConfigInfo,
    #[serde(default)]
    pub debug: bool,
    #[serde(default)]
    pub token_secret: String,
    #[serde(default)]
    pub session_length: u64,
    #[serde(default = "default_port")]
    pub port: u16,
    #[serde(default)]
    #[serde(alias = "DATABASE_URL")]
    pub database_url: String,
    #[serde(default)]
    pub ldap: LDAP,
}

impl Configuration {
    pub fn new(location: PathBuf) -> Result<Self, Report<AppError>> {
        let s = Config::builder()
            .add_source(File::from(location.clone()))
            .add_source(Environment::default())
            .set_override("config.location", location.to_str().unwrap_or("default"))
            .change_context(AppError)?
            .build()
            .change_context(AppError)?;

        let settings = s.try_deserialize().change_context(AppError)?;

        Ok(settings)
    }

    pub fn check(&self) -> Result<(), Report<AppError>> {
        // check if elevated LDAP user is set if activated
        if self.ldap.elevated_search {
            if self.ldap.elevated_user_id == "" {
                return Err(Report::new(AppError).attach_printable(
                    "LDAP: Search with elevated user is activated but no user is configured",
                ));
            }

            if self.ldap.elevated_user_pw == "" {
                return Err(Report::new(AppError).attach_printable(
                    "LDAP: Search with elevated user is activated but no password is configured",
                ));
            }
        }
        Ok(())
    }
}
